
from typing import Any, List, Tuple

class GraderCondition(dict):
    """Convience class to represent the a grader condition for an exercise.
    
    Example:
    
    GraderCondition(
      x = x,
      message = message,
      correct = correct,
      type = type
    )
    """
    def __init__(self, *args, **kwargs):
      # TODO validation of kwargs to only accept the valid names
      super(GraderCondition, self).__init__(kwargs)

def python_condition(x: Any, message: str, correct: bool, type: str = "value") -> dict:
  """Return the proper structure for a particular type of condition."""
  # Note: we don't use the type field from `gradethis::condition` yet, so assumes value
  # TODO think about whether we allow passing of a function (lambda or regular)
  return GraderCondition(x = x, message = message, correct = correct, type = type)

def python_pass_if(x: Any, message = "") -> dict:
  """Return a pass condition."""
  return python_condition(x, message, correct = True)

def python_fail_if(x: Any, message = "") -> dict:
  """Return a fail condition."""
  return python_condition(x, message, correct = False)



"""
This module is used to check if a user's code is equal to solution code and
note when the two differ.
"""

# parsing
import parser
import ast
import asttokens

# formatting
from .formatters import formatted
# checking functions
from .check_functions import standardize_arguments

# misc
from itertools import zip_longest
from typing import Optional, Dict, Any

def new_parent_source(atok: asttokens.ASTTokens, tree: Any, last_parent: str) -> str:
    """Given the token-marked ast, the current tree, and the last parent source,
    return a new last parent source for facilitating feedback on problematic
    node.

    Parameters
    ----------
    atok : asttokens.ASTTokens
        the token-marked AST
    tree : Any
        the node we want to extract the source text for
    last_parent : str
        the nearest parent that can be converted to source text

    Returns
    -------
    str
        the new last_parent or the same
    """
    # attempt to get a new parent source text for feedback
    parent_source = atok.get_text(tree)
    # only set the last_parent if:
    # - it is a compound ast, and
    # - we can get a source text
    if isinstance(tree, ast.AST) and len(tree._fields) > 1 and parent_source != "":
        last_parent = parent_source
    return last_parent

def check_children(u_atok: asttokens.ASTTokens, 
                   s_atok: asttokens.ASTTokens,
                   left: ast.AST, 
                   right: ast.AST, 
                   line_info: Optional[Dict[str, int]] = None,
                   last_parent: str = ""):
    """Checks children of two asts by iterating their fields

    Parameters
    ----------
    u_atok : asttokens.ASTTokens
        the token-marked AST for the student code
    s_atok : asttokens.ASTTokens
        the token-marked AST for the solution code
    left : ast.AST
        the student code
    right : ast.AST
        the solution code
    line_info :  Optional[Dict[str, int]], optional
        holds line information about a particular node, by default None
    last_parent : str, optional
        the nearest parent that can be converted to source text, by default ""
    """
    lf, rf = ast.iter_fields(left), ast.iter_fields(right)
    # iterate through the children of both ASTs
    for left_field, right_field in zip_longest(lf, rf, fillvalue=""):
        left_name, left_values = left_field
        right_name, right_values = right_field
        # attempt to get a new parent source text for feedback
        # check that the name of the AST nodes match
        compare_node(left_name, right_name, line_info, last_parent)
        # recurse on values
        compare_ast(u_atok, s_atok, left_values, right_values, line_info, last_parent)

def compare_ast(u_atok: asttokens.ASTTokens, 
                s_atok: asttokens.ASTTokens,
                left: ast.AST, 
                right: ast.AST, 
                line_info: Optional[Dict[str, int]] = None,
                last_parent: str = "") -> None:
    """Compare two abstract syntax trees. Raise AssertionError as soon as they differ.

    Parameters
    ----------
    atok : asttokens.ASTTokens
        the token-marked AST for the student code
    left : ast.AST
        the student code
    right : ast.AST
        the solution code
    line_info :  Optional[Dict[str, int]], optional
        holds line information about a particular node, by default None
    last_parent : str, optional
        the nearest parent that can be converted to source text, by default ""
    """
    # to hold line information like line number, and original source
    line_info = {} if line_info is None else line_info
    # check types first
    compare_node_type(left, right, line_info, last_parent)

    # the check AST, list of Expr, or a core data type
    if isinstance(left, ast.AST):
        # store line number for feedback
        line_info["left"] = getattr(
            left, "lineno", line_info.get("left", 1)
        )
        line_info["right"] = getattr(
            right, "lineno", line_info.get("right", 1)
        )
        # for ast.Call we will raise error when there is either a problem 
        # with the function or the arguments do not match after standardization
        if isinstance(left, ast.Call):
            # attempt to get a new parent source text for feedback
            last_parent = new_parent_source(u_atok, left, last_parent)
            # check that the standardized left and right Calls are the same
            check_functions(u_atok, s_atok, left, right, line_info, last_parent)
        else:
            check_children(u_atok, s_atok, left, right, line_info, last_parent)
    elif isinstance(left, list):
        for left_child, right_child in zip_longest(left, right, fillvalue=""):
            # recurse on [Expr, ...]
            # attempt to get a new parent source text for feedback
            last_parent = new_parent_source(u_atok, left_child, last_parent)
            compare_ast(u_atok, s_atok, left_child, right_child, line_info, last_parent)
    else:
        compare_node(left, right, line_info, last_parent)

# TODO document
def check_functions(u_atok, s_atok, left, right, line_info, last_parent):
    # standardize the left and right tree
    ls = standardize_arguments(left, u_atok.get_text(u_atok.tree))
    rs = standardize_arguments(right, s_atok.get_text(s_atok.tree))
    msg = "I expected {}, but you wrote {} in {} at line {}."
    msg_args = (
        formatted(rs),
        formatted(ls),
        last_parent,
        line_info.get("left"),
    )
    # if we don't have any arguments simply compare the two nodes
    if len(ls.keywords) == 0:
        check_children(u_atok, s_atok, left, right, line_info, last_parent)
    for l, r in zip_longest(ls.keywords, rs.keywords, fillvalue=""):
        # TODO move this to wrong_value in message_generators
        msg = "I expected {}, but what you wrote was interpreted as {} at line {}."
        msg_args = (
            formatted(rs),
            formatted(ls),
            line_info.get("left"),
        )
        assert formatted(l.value) == formatted(r.value), msg.format(*msg_args)
    
def compare_node_type(
        left: Any, 
        right: Any, 
        line_info: Dict[str, int],
        last_parent: str
    ) -> None:
    """Compare two ASTs' types and raise an exception with the line number if different.

    Parameters
    ----------
    left : Any
        an ast.Node, a custom type, or a python data type
    right : Any
        an ast.Node, a custom type, or a python data type
    line_info : Dict[str, int]
        holds line information about a particular node
    last_parent : str
        the nearest parent that can be converted to source text
    """
    # TODO find a more maintaible way to handle the string setup for feedback here
    # str stands for "empty", so if user code is missing something, present 
    # a message about missing expectation  
    # print(left, right)
    msg = "I expected {} at line {}."
    msg_args = (
        formatted(right), 
        line_info.get("left")
    )
    if isinstance(left, str) and not isinstance(right, str):
        # TODO yes this is code duplication from default message but refactor in
        # feedback.py
        msg = "I expected {} at line {}."
        msg_args = (
            formatted(right), 
            line_info.get("left")
        )
    # otherwise, present a message about unexpected user's AST
    elif not isinstance(left, str):
        # if we do expect a certain AST, point out that expectation
        if not isinstance(right, str):
            msg = "I expected {}, but what you wrote was interpreted as {} at line {}."
            msg_args = (
                formatted(right),
                formatted(left),
                line_info.get("left"),
            )
        # otherwise, just state what was not expected
        else:
            msg = "I did not expect {} at line {}."
            msg_args = (
                formatted(left), 
                line_info.get("left")
            )
    assert type(left) == type(right), msg.format(*msg_args)

def compare_node(
        left: Any, 
        right: Any, 
        line_info: Dict[str, int],
        last_parent: str
    ):
    """Compare two objects of the same type and raise an exception with feedback if nodes differ.

    Parameters
    ----------
    left : Any
        an ast.Node, a custom type, or a python data type
    right : Any
        an ast.Node, a custom type, or a python data type
    line_info : Dict[str, int]
        holds line information about a particular node
    last_parent : str
        the nearest parent that can be converted to source text
    """
    msg = "I expected {}, but you wrote {} in {} at line {}."
    msg_args = (
        formatted(right),
        formatted(left),
        last_parent,
        line_info.get("left"),
    )

    assert left == right, msg.format(*msg_args)

def grade_code(student_code: str, solution_code: str):
    """Checks user and solution code and prints a message if they differ

    Parameters
    ----------
    student_code : str
        the user source code
    solution_code : str
        the solution source code

    Returns
    -------
    None
        If there are no issues
    str
        If user code differs from solution
    """
    try:
        # TODO: make parser more pluggable for 3.8 ast parser or other potential parsers like antlr
        # with the contract being it needs a similar get_text for getting
        # source node back
        student = asttokens.ASTTokens(student_code, parse=True)
        solution = asttokens.ASTTokens(solution_code, parse=True)
        # set parent node of child nodes for better feedback
        u_atok = student
        s_atok = solution
        last_parent = u_atok.get_text(next(n for n in ast.walk(u_atok.tree)))
        compare_ast(u_atok, s_atok, student.tree, solution.tree, last_parent = last_parent)
    except AssertionError as e:
        return str(e) # back to either the python_grader or python_grade_learnr
    except Exception as e:
        return "There was a problem checking user or solution code: {}".format(e)

if __name__ == "__main__":
    # formatting
    from formatters import formatted
    # checking functions
    from check_functions import standardize_arguments
    # add any libraries you might need
    from math import sqrt, log
    # TODO move to testing class
    # these are the non-verbose cases
    test_cases = [
        # different core types
        ("\"1\"", "1", "I expected 1, but what you wrote was interpreted as \"1\" at line 1."),
        ("1", "\"1\"", "I expected \"1\", but what you wrote was interpreted as 1 at line 1."),
        ("[1]", "\"1\"", "I expected \"1\", but what you wrote was interpreted as [1] at line 1."),
        ("True", "\"1\"", "I expected \"1\", but what you wrote was interpreted as True at line 1."),
        ("False", "\"1\"", "I expected \"1\", but what you wrote was interpreted as False at line 1."),
        # str vs str
        ("\"not hello\"", "\"hello\"", "I expected \"hello\", but you wrote \"not hello\" in \"not hello\" at line 1."),
        # list vs list
        ("[2]", "[]", "I did not expect 2 at line 1."),
        ("[1,2]", "[1]", "I did not expect 2 at line 1."),
        ("[1]", "[1,2]", "I expected 2 at line 1."),
        # expr vs something else
        ("1 + 1", "1", "I expected 1, but what you wrote was interpreted as 1 + 1 at line 1."),
        ("-1", "1", "I expected 1, but what you wrote was interpreted as -1 at line 1."),
        # functions
        # TODO this is an odd one, maybe we could just revert to original source text for these?
        ("2 + sum([1,2])", "2 + sum([1,1])", "I expected sum(iterable=[1, 1], start=0), but what you wrote was interpreted as sum(iterable=[1, 2], start=0) at line 1."),
        # TODO handle functions where you don't have function signature like math.sqrt / math.log
        ("sqrt(log(2))", "sqrt(log(1))", "I expected 1, but you wrote 2 in log(2) at line 1."),
        ("def foo(a, b=1): pass; foo(2)", "def foo(a, b=1): pass; foo(1)", "I expected foo(a=1, b=1), but what you wrote was interpreted as foo(a=2, b=1) at line 1."),
        ("def foo(a, b=1): pass; foo(a=2)", "def foo(a, b=1): pass; foo(1)", "I expected foo(a=1, b=1), but what you wrote was interpreted as foo(a=2, b=1) at line 1."),
        ("def foo(a, b=1): pass; foo(a=[2])", "def foo(a, b=1): pass; foo(2, 2)", "I expected foo(a=2, b=2), but what you wrote was interpreted as foo(a=[2], b=1) at line 1."),
        ("def foo(a, b=1): pass; foo(a=\"2\", b=2)", "def foo(a, b=1): pass; foo(2, 2)", "I expected foo(a=2, b=2), but what you wrote was interpreted as foo(a=\"2\", b=2) at line 1."),
        ("def head(n=5): pass; head(12)", "def head(n=5): pass; head(n=10)", "I expected head(n=10), but what you wrote was interpreted as head(n=12) at line 1."),
    ]
    for t in test_cases:
        # print("~~~")
        # print("user code:\n{}\n".format(t[0]))
        # print("solution code:\n{}\n".format(t[1]))
        message = grade_code(t[0], t[1])
        if message != t[2]:
            raise ValueError(
                "Failed test case!\nUser:{}\nSolution:{}\nExpected:{}\nGot:{}".format(
                    t[0],
                    t[1],
                    t[2],
                    message
                )
            )
    print("All tests passed! :)")



"""
This module contains functions to faciliate checking Python code or output.
"""
from typing import Union

from .grade_result import python_grade_result
from .grade_code import grade_code
from .conditions import GraderCondition
from .feedback import praise, encourage
from .utils import parse_code

def graded(result: Union[str, dict], condition: GraderCondition):
  # return a dict for feedback
  if (result and condition):
    # correct
    if condition['correct']:
      return dict(
        message = "{} {}".format(praise(), condition['message']), 
        correct = True,
        type = "success"
      )
    # incorrect
    return dict(
      message = "{} {}".format(encourage(), condition['message']),
      correct = False,
      type = "error"
    )
  # if none of the conditions matched, return incorrect by default
  elif not result and not condition:
    return dict(
      message = "{}".format(encourage()), 
      correct = False, 
      type = "error"
    )
  # if there were fail_ifs and none matched, return success by default
  elif result and not condition:
    return dict(
      message = "{}".format(praise()),
      correct = True,
      type = "success"
    )
  else:
    return dict(
      message = "{} {}".format(encourage(), condition['message']), 
      correct = False, 
      type = "error"
    )


def grade(*check_code: GraderCondition,
          user_code: str = None, 
          solution_code: str = None) -> dict:
  """Python standalone verson of `python_grade_learnr`. This function does
  two things:
  - Do static code grading (AST) if both user and solution code is provided
    before the result grading.
  - Do result grading based on code output and GraderCondition(s)

  Parameters
  ----------
  check_code: *GraderCondition
      a variable number of GraderCondition objects
  user_code : str, optional
      user source code, by default None
  solution_code : str, optional
      solution source code, by default None

  Returns
  -------
  dict
      a feedback dict:
      {
        message: str,
        correct=True|False,
        type="auto|success|info|warning|error|custom",
        location="append|prepend|replace"
      }
  """
  # check if there is user_code
  if user_code and "".join(user_code) == "":
    return dict(
      message = "I didn't receive your code. Did you write any?",
      correct = False,
      type = "error"
    )
  # if there is check code and solution code, check if there is a solution code
  if (check_code and solution_code) and "".join(solution_code) == "":
    return dict(
      message = "No solution is provided for this exercise.",
      correct = True,
      type = "info"
    )
  
  # TODO input scrubbing for malicious code
  # parse user, and solution code
  solution_code = parse_code(solution_code)
  user_code_source = parse_code(user_code)

  # print('user_code_source', user_code_source)
  # print('solution_code', solution_code)
  if solution_code is not None:
    # do static checks on code if the solution code is provided
    graded_code = grade_code(user_code_source, solution_code)
    if graded_code is not None:
      return dict(
        message = "{}".format(graded_code),
        correct = False,
        type = "error", 
      )

  # evaluate exercise and check code output
  try:
    # evaluate user code so that we can compare to expected
    user_result = eval(user_code_source)
  except Exception as e:
    # TODO somehow trickle up the specific error message?
    return dict(
      message="Error occured while checking the submission. {e}".format(e),
      correct = False, 
      type = "warning"
    )

  # grade python_pass_if/fail_if conditions against user's code output
  result, condition = python_grade_result(*check_code, user_result = user_result)
  # return a graded dict
  return graded(result, condition)





"""
Module contains a generic AST node formatter and allows for custom formatters
for whichever data type you need.

To define a node formatter simply define another `_` function, register its 
type for the generic `formatted` function, and define the string to return:

```
@formatted.register(<your type>)
def _(node):
    <process your node>
    return "<your string>"
```

Alternatively, you can also call the `format_node` function to register your
function using one of two ways:

1. `format_node(int, <"a string">)`

2. `format_node(int, <a function or lambda that formats and returns string>)`

The basic data types and ast types have been supported but can be changed in
the respective register function definition, or overriden with `format_node`.
"""

import ast
from functools import singledispatch

@singledispatch
def formatted(node):
    """Generic function to return a formatted representation of an AST node"""
    return "{}".format(type(node).__name__.lower())

def format_node(node_type, string_format):
    """Registers a format function for a give node type
    
    The node_type must be a valid data type in Python.
    The string_format can be either a str or a function / lambda.
    """
    @formatted.register(node_type)
    def _(node):
        if callable(string_format):
            return string_format(node)
        else:
            return string_format

# binary operators
format_node(ast.Add, "+")
format_node(ast.Sub, "-")
format_node(ast.Mult, "*")
format_node(ast.MatMult, "@")
format_node(ast.Div, "/")
format_node(ast.FloorDiv, "//")
format_node(ast.Mod, "%")
format_node(ast.Pow, "**")
format_node(ast.LShift, "<<")
format_node(ast.RShift, ">>")
format_node(ast.BitOr, "|")
format_node(ast.BitXor, "^")
format_node(ast.BitAnd, "&")
# unary operators
format_node(ast.UAdd, "+")
format_node(ast.USub, "-")
format_node(ast.Invert, "~")
# bool operators
format_node(ast.And, "and")
format_node(ast.Or, "or")
format_node(ast.Not, "not")
# compare operators
format_node(ast.Eq, "==")
format_node(ast.NotEq, "!=")
format_node(ast.Lt, "<")
format_node(ast.LtE, "<=")
format_node(ast.Gt, ">")
format_node(ast.GtE, ">=")
format_node(ast.Is, "is")
format_node(ast.IsNot, "is not")
format_node(ast.In, "in")
format_node(ast.NotIn, "not in")

# core types

@formatted.register(str)
def _(el):
    """Formatting function for an str"""
    return "\"{}\"".format(el)

@formatted.register(int)
def _(el):
    """Formatting function for an int"""
    return el

@formatted.register(ast.List)
def _(node):
    """Formatting function for an ast.List"""
    return [formatted(n) for n in node.elts]

@formatted.register(ast.Tuple)
def _(node):
    """Formatting function for an ast.Tuple"""
    return tuple([formatted(n) for n in node.elts])

@formatted.register(ast.Set)
def _(node):
    """Formatting function for an ast.Set"""
    return {formatted(n) for n in node.elts}

@formatted.register(ast.Dict)
def _(node):
    """Formatting function for an ast.Dict"""
    adict = {}
    for k, v in zip(node.keys, node.values):
        if isinstance(k, ast.Str):
            adict[k.s] = formatted(v)
        else:
            adict[formatted(k)] = formatted(v)
    return adict

@formatted.register(ast.Expr)
def _(node):
    """Formatting function for an ast.Expr"""
    return "an expression"

@formatted.register(ast.arg)
def _(node):
    return node.arg

@formatted.register(ast.arguments)
def _(node):
    defaults = [formatted(d) for d in node.defaults]
    len_defaults = len(defaults)
    args = [formatted(a) for a in node.args[:len_defaults-1]]
    rem_args = [formatted(a) for a in node.args[len_defaults-1:]]
    map_args_defaults = ["{}={}".format(a, d) for a, d in zip(rem_args, defaults)]
    return  ", ".join(args + map_args_defaults)

@formatted.register(ast.Lambda)
def _(node):
    args = formatted(getattr(node, 'args'))
    return "lambda {}: {}".format(args, formatted(node.body))

@formatted.register(ast.Call)
def _(node):
    """Formatting function for an ast.Call"""
    if isinstance(node.func, ast.Attribute):
        func_name = getattr(node.func, "attr") 
    else:
        func_name = node.func.id
    # grab args and keywords
    args = [str(formatted(a)) for a in node.args]
    keywords = ["{}={}".format(k.arg, formatted(k.value)) for k in node.keywords]
    # return call like it would appear
    return "{}({})".format(func_name, ", ".join(args + keywords))

@formatted.register(ast.keyword)
def _(node):
    """Formatting function for an ast.keyword"""
    return getattr(node, "arg")

@formatted.register(ast.BinOp)
def _(node):
    """Formatting function for an ast.BinOp"""
    return "{} {} {}".format(formatted(node.left), formatted(node.op), formatted(node.right))

@formatted.register(ast.UnaryOp)
def _(node):
    """Formatting function for an ast.UnaryOp"""
    return "{}{}".format(formatted(node.op), formatted(node.operand))

@formatted.register(ast.Attribute)
def _(node):
    """Formatting function for an ast.Attribute"""
    return "\"{}\" on {}".format(getattr(node, "attr"), formatted(getattr(node, "value")))

@formatted.register(ast.Name)
def _(node):
    """Formatting function for an ast.Name"""
    return getattr(node, "id")

@formatted.register(ast.Num)
def _(node):
    """Formatting function for an ast.Num"""
    return getattr(node, "n")

@formatted.register(ast.Str)
def _(node):
    """Formatting function for an ast.Str"""
    return "\"{}\"".format(getattr(node, "s"))

@formatted.register(ast.NameConstant)
def _(node):
    """Formatting function for True/False"""
    return getattr(node, "value")

def test_format(node):
    print(formatted(node))



"""
Module with useful functions.
"""

from typing import Union, List
import ast

def parse_code(input: Union[str, List[str]]) -> str:
    """Tries to parse code represented as string or list of strings

    Parameters
    ----------
    input : Union[str, List[str]]
        either a str or a list of str

    Returns
    -------
    str
        the formatted string

    Raises
    ------
    SyntaxError
        if there are any parsing issues
    """
    if input is None:
        return input
    try:
        simple = "".join(input)
        ast.parse(simple)
        return simple
    except SyntaxError as e:
        if "EOF" in str(e):
            return "\n".join(input)
        else:
            raise SyntaxError("Problem parsing your code!")




"""This module contains utilities for generating a friendly message for feedback"""
import random

def praise() -> str:
  """Returns a random praise message"""
  return random.choice([
      "Absolutely fabulous!",
      "Amazing!",
      "Awesome!",
      "Beautiful!",
      "Bravo!",
      "Cool job!",
      "Delightful!",
      "Excellent!",
      "Fantastic!",
      "Great work!",
      "I couldn't have done it better myself.",
      "Impressive work!",
      "Lovely job!",
      "Magnificent!",
      "Nice job!",
      "Out of this world!",
      "Resplendent!",
      "Smashing!",
      "Someone knows what they're doing :)",
      "Spectacular job!",
      "Splendid!",
      "Success!",
      "Super job!",
      "Superb work!",
      "Swell job!",
      "Terrific!",
      "That's a first-class answer!",
      "That's glorious!",
      "That's marvelous!",
      "Very good!",
      "Well done!",
      "What first-rate work!",
      "Wicked smaht!",
      "Wonderful!",
      "You aced it!",
      "You rock!",
      "You should be proud.",
      ":)"
    ]
  )

# TODO move this to feedback.py
def encourage() -> str:
  """Returns a random encouragement message"""
  return random.choice([
      "Please try again.",
      "Give it another try.",
      "Let's try it again.",
      "Try it again; next time's the charm!",
      "Don't give up now, try it one more time.",
      "But no need to fret, try it again.",
      "Try it again. I have a good feeling about this.",
      "Try it again. You get better each time.",
      "Try it again. Perseverence is the key to success.",
      "That's okay: you learn more from mistakes than successes. Let's do it one more time."
    ]
  )


"""
This module is used to check python code output.
"""

from typing import Any, Union, List, Tuple
import pandas as pd

from .conditions import GraderCondition

def python_compare_output(user_output: Any, expected_output: Any) -> bool:
  """Return whether the user output and expected output match"""
  if type(user_output) != type(expected_output):
    return False
  elif isinstance(user_output, pd.DataFrame) and isinstance(expected_output, pd.DataFrame):
      return user_output.equals(expected_output)
  else:
    return user_output == expected_output
  
def python_grade_conditions(*conditions: GraderCondition, user_result: Any = None) -> Tuple[bool, GraderCondition]:
  """Goes through all conditions (python_pass_if, python_fail_if) and
  returns the first condition that comes True.

  Parameters
  ----------
  conditions : List[GraderCondition]
      
  user_result : Any
      [description]

  Returns
  -------
  Tuple[bool, GraderCondition]
      bool is True if user_result is correct, else False
      GraderCondition that matched with the user_result
  """
  # TODO handle the case where you might only have fail_ifs but they don't match (issue #4)
  result, condition = False,  None
  for cond in conditions:
    condition = cond
    result = python_compare_output(cond['x'], user_result)
    if result:
      return result, condition
  # If there is at least one pass_if() condition, then default to an incorrect grade;
  if len([c for c in conditions if c['correct']]) != 0:
    return False, None
  # otherwise if we only have fail_ifs, we default to a correct grade 
  # see https://github.com/rstudio-education/gradethis/issues/118
  elif len([c for c in conditions if not c['correct']]) != 0:
    return True, None
  return False, None

# TODO add the other kwargs for this?
# glue_correct = getOption("gradethis_glue_correct"),
# glue_incorrect = getOption("gradethis_glue_incorrect")
def python_grade_result(*conditions: List[GraderCondition], user_result: Any = None) -> Union[dict, str]:
  """This function checks the user's code output against the list of conditions.

  Parameters
  ----------
  conditions: *GraderCondition
      a variable number of GraderCondition objects
  user_result : Any, optional
      user's code output, by default None

  Returns
  -------
  Union[dict, str]
      dict, if there are no issues
      str, an error message if GraderCondition(s) is not supplied

  Raises
  ------
  Exception
      if GraderConditions are not passed in for conditions
  """
  if conditions == None or len([c for c in conditions if isinstance(c, GraderCondition)]) == 0:
    raise Exception(
      "At least one condition object (e.g., `python_pass_if()`, "
      "`python_fail_if()`, `python_condition()`) must be provided to"
      "`python_grade_result()`"
    )
  if user_result is not None:
    return python_grade_conditions(*conditions, user_result=user_result)
  return conditions



"""
This is the `learnr` version of `python_grader` module and contains
the `python_grade_learnr` function called by the `gradethispython` R wrapper
package
"""
import parser
from typing import Any, Union, List, Tuple

from .grade_result import python_grade_result
from .grade_code import grade_code
from .conditions import GraderCondition
from .feedback import praise, encourage
from .utils import parse_code

def graded_learnr(result: Union[str, dict], condition: GraderCondition):
  if (result and condition):
    # correct
    if condition['correct']:
      return dict(
        message = "{} {}".format(praise(), condition['message']), 
        correct = True,
        type = "success", 
        location = "append"
      )
    # incorrect
    return dict(
      message = "{} {}".format(encourage(), condition['message']),
      correct = False,
      type = "error", 
      location = "append"
    )
  # if there was none of the conditions matched, return error by default
  elif not result and not condition:
    return dict(
      message = "{}".format(encourage()), 
      correct = False, 
      type = "error", 
      location = "append"
    )
  # if there were fail_ifs and none matched, return success by default
  elif result and not condition:
    return dict(
      message = "{}".format(praise()),
      correct = True,
      type = "success", 
      location = "append"
    )
  else:
    return dict(
      message = "{} {}".format(encourage(), condition['message']), 
      correct = False, 
      type = "error", 
      location = "append"
    )

def python_grade_learnr(label: str = None,
                        solution_code: str = None,
                        user_code: str = None,
                        check_code: List[str] = None,
                        envir_result: dict = None,
                        evaluate_result: List[str] = None,
                        envir_prep: dict = None,
                        last_value: Any = None,
                        **kwargs) -> dict:
  """This function mirrors the `grade_learnr` function from {gradethis} package so that
  we can check Python exercises. This function does two things:
  - Do static code grading (AST) if both user and solution code is provided
    before the result grading.
  - Do result grading based on code output and GraderCondition(s)

  Parameters
  ----------
  label : str, optional
      exercise label, by default None
  solution_code : str, optional
      solution source code, by default None
  user_code : str, optional
      user source code, by default None
  check_code : List[str], optional
      checking code, by default None
  envir_result : dict, optional
      environment containing execution results, by default None
  evaluate_result : List[str], optional
      evaluation results, by default None
  envir_prep : dict, optional
      environment associated with exercise, by default None
  last_value : Any, optional
      the last value of exercise, by default None

  Returns
  -------
  dict
      a feedback dict:
      {
        message = str,
        correct = True|False,
        type = "auto|success|info|warning|error|custom",
        location = "append|prepend|replace"
      }
  """
  # check if there is user_code
  if user_code and "".join(user_code) == "":
    return dict(
      message="I didn't receive your code. Did you write any?",
      correct=False,
      type="error",
      location="append"
    )
  # if there is check code and solution code, check if there is a solution code
  if (check_code and solution_code) and "".join(solution_code) == "":
    return dict(
      message="No solution is provided for this exercise.",
      correct=True,
      type="info",
      location="append"
    )

  # TODO input scrubbing for malicious code
  # parse user, check and solution code
  solution_code = parse_code(solution_code)
  check_code_source = parse_code(check_code)
  user_code_source = parse_code(user_code)

  # do static checks on code if the solution code is provided
  graded_code = grade_code(user_code_source, solution_code)
  if graded_code is not None:
    return dict(
      message = "{}".format(graded_code),
      correct = False,
      type = "error", 
      location = "append"
    )

  # evaluate exercise and check code output
  try:
    # Note: because Python is eager evaluation, we already have introduced
    # the `r` object in the current scope when entering this function
    # evaluate check code so that expected output is ready
    check_code_conditions = eval(check_code_source, {}, r)
    # for e.g. did knitr already execute result and it's somewhere in the `r`?
    # evaluate user code so that we can compare to expected
    user_result = eval(user_code_source, {}, r)
    # grade python_pass_if/fail_if conditions against user's code output
    result, condition = python_grade_result(check_code_conditions, user_result = user_result)
  except Exception as e:
    # TODO somehow trickle up the specific error message?
    return dict(
      message="Error occured while checking the submission. {e}".format(e),
      correct=False,
      type="warning",
      location="append"
    )
  # return a dict for learnr to process for feedback
  return graded_learnr(result, condition)




import unittest

from pygradethis.conditions import *
from pygradethis.python_grader import *

class PythonGraderTest(unittest.TestCase):

    # Grade result -------------------------------------

    def test_pass_if_correct(self):
        result = grade(
            python_pass_if(2, "Woah, nice!"),
            user_code="2", 
            solution_code="2", 
        )
        self.assertTrue(result['correct'])
        self.assertEqual(result['type'], "success")

    def test_pass_if_incorrect(self):
        result = grade(
            python_pass_if(2, "Woah, nice!"),
            user_code="1", 
            solution_code="2", 
        )
        self.assertFalse(result['correct'])
        self.assertEqual(result['type'], "error")
    
    def test_multiple_pass_if_correct(self):
        result = grade(
            python_pass_if(2, "Woah, nice!"),
            python_pass_if(3, "Piece of cake!"),
            user_code="3",
        )
        self.assertTrue(result['correct'])
        self.assertEqual(result['type'], "success")
    
    def test_multiple_pass_if_incorrect(self):
        result = grade(
            python_pass_if(2, "Woah, nice!"),
            python_pass_if(3, "Piece of cake!"),
            user_code="1",
        )
        self.assertFalse(result['correct'])
        self.assertEqual(result['type'], "error")

    def test_pass_if_fail_if_correct(self):
        result = grade(
            python_pass_if(2, "Noice!"),
            python_fail_if(1, "We do not want 1."),
            user_code="2",
            solution_code="2", 
        )
        self.assertTrue(result['correct'])
        self.assertEqual(result['type'], "success")

    def test_pass_if_fail_if_incorrect(self):
        result = grade(
            python_pass_if(2, "Noice!"),
            python_fail_if(1, "We do not want 1."),
            user_code="1",
            solution_code="2", 
        )
        self.assertFalse(result['correct'])
        self.assertEqual(result['type'], "error")

    def test_fail_if_correct(self):
        # single fail_if; if user don't match any fail_ifs, count as correct by default
        result = grade(
            python_fail_if(1, "We wanted 1."),
            user_code="3",
        )
        self.assertTrue(result['correct'])
        self.assertEqual(result['type'], "success")
    
    def test_fail_if_incorrect(self):
        result = grade(
            python_fail_if(1, "We wanted 1."),
            user_code="1",
        )
        self.assertFalse(result['correct'])
        self.assertEqual(result['type'], "error")

    def test_fail_ifs_correct(self):
        # multiple fail_ifs; if user don't match any fail_ifs, count as correct by default
        result = grade(
            python_fail_if(1, "We do not want 1."),
            python_fail_if(2, "We do not want 2."),
            user_code="3",
        )
        self.assertTrue(result['correct'])
        self.assertEqual(result['type'], "success")
    
    def test_fail_ifs_incorrect(self):
        result = grade(
            python_fail_if(1, "We do not want 1."),
            python_fail_if(2, "We do not want 2."),
            user_code="2",
        )
        self.assertFalse(result['correct'])
        self.assertEqual(result['type'], "error")
    
    # Grade code -------------------------------------
    # TODO add static code check examples
    

if __name__ == "__main__":
    unittest.main()


"""
This module contains functions to check function calls, standardize arguments,
and return standardized call.
"""
import ast
import asttokens
import inspect
import builtins

from .formatters import formatted
from math import sqrt, log

from itertools import zip_longest

def standardize_arguments(call: ast.Call, source_code: str) -> ast.Call:
    """This will standardize the function calls for a function Call and return
    the modified Call.
    
    Returns None if there are no issues with the function call.
    
    Otherwise, an error message is returned if the user code can't be parsed, 
    if there is a problem not finding variables in environment, or if arguments 
    cannot be standardized.

    Parameters
    ----------
    call : ast.Call
        the call to standardize arguments
    source_code : str
        the source code in which the call belongs to

    Returns
    -------
    ast.Call
        the standardized call

    Raises
    ------
    TypeError
        when call itself is invalid (e.g. wrong argument name)
    NameError
        when for e.g. the function name doesn't exist in environment
    Exception
        when other Exceptions occur, silently ignore for now
    """
    try:
        # 1) introduce function call in environment
        # first pass: is it legal Python code?
        # for e.g., positional args should always be before keywords args
        # exec will catch these issues
        exec(source_code)
        # 2) construct a environment containing encompassing both global and locals, 
        # overwriting the globals with locals, and builtins.
        envir = dict(globals(), **locals(), **builtins.__dict__)
        # 3) grab the live function from the environment
        # if we're calling a function on an object, the function info has
        # to be extracted from an ast.Attribute
        if isinstance(call.func, ast.Attribute):
            func_name = getattr(call.func, "attr") 
            # the object's name 
            obj = call.func.value.id
            # the live function from the environment associated with object
            # e.g. the `df` in df.head() is a pd.DataFrame so must get the
            # `head` associated with that class.
            live_func = getattr(envir[obj], func_name)
        else:
            # if we're calling a function not associated with an object
            # just grab function from environment using its name
            func_name = call.func.id
            live_func = envir[func_name]

        # 4) collect the arguments passed
        # construct keyword args mapping
        kwargs = {a.arg:a.value for a in call.keywords}
        
        # 5) get the formal arguments for function
        try:
            # Note: this will raise ValueError if inspect cannot retrieve signature
            # which can happen for some builtins like `print`, where underlying C
            # code does not provide any metadata about its signature.
            sig = inspect.signature(live_func)
        except ValueError: 
            # if we can't get a signature just return call for normal Call
            # checking flow
            return call

        # 6) unpack args and kwargs and attempt to standardize argument calls
        # returns: https://docs.python.org/3.6/library/inspect.html#inspect.BoundArguments
        partial_args = sig.bind(*call.args, **kwargs)
        partial_args.apply_defaults()

        # 7) return a modified AST representing the standardized call
        # we do this by updating Call.keywords and reset the args
        # TODO handle functions where you don't have keywords: import math; math.sqrt(math.log(1))
        new_keywords = []
        for k, v in partial_args.arguments.items():
            new_keywords.append(ast.keyword(arg=k, value=v))
        call.args = []
        call.keywords = new_keywords
        return call
    except TypeError as e:
        raise AssertionError(str(e).capitalize())
    except NameError as e:
        raise AssertionError(str(e).capitalize())
    except Exception as e:
        pass



"""
This module holds functions for common mistakes and generating messages for
feedback to the student.
"""

# TODO: use these from `python_grader`, `python_grade_learnr`, and `grade_code`

def wrong_value(this, that):

    # def foo(a, b=1): pass
    
    # foo(1)
    # foo(2)

    # "h(1), I expected 2 where you wrote 1."
    # "I expected {that} where you wrote {this}."
    pass


def duplicate_name(this_call, this_name):

    # def foo(a, b=1): pass

    # foo(a=1, a=2)
    # SyntaxError: keyword argument repeated

    # "You passed multiple arguments named {this_name} to {this_call}, which will cause "
    # "an error. Check your spelling, or remove one of the arguments."
    pass

def wrong_call(this, that):
    # def foo(a, b=1): pass

    # bar(1)
    # foo(1)

    # "I expected you to call {that} where you called {this}."
    pass

def bad_argument_name(this, this_name):
    # def foo(a, b=1): pass
    # foo(x=1)

    # TypeError: foo() got an unexpected keyword argument 'x'

    # {this} got an unexpected keyword argument {this_name}
    pass


def missing_argument(this, that_name):
    # def foo(a, b=1): pass
    # foo() missing 1 required positional argument: 'a'

    # "Your call to {this} should include {that_name} ",
    # "You may have misspelled an argument name, ",
    # "or left out an important argument."
    pass


def surplus_argument(this, this_arg):

    # def foo(a, b=1): pass
    # foo(x=1) => TypeError: foo() got an unexpected keyword argument 'x'

    # "I did not expect your call to {this} to ",
    # "include {this_arg}. You ",
    # "may have included an unnecessary argument, or you ",
    # "may have left out or misspelled an important ",
    # "argument name."
    pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pygradethis",
    version="0.1.0",
    author="Nischal Shrestha",
    author_email="nsrocker92@gmail.com",
    description="Python autograder to facilitate code output and static code checking.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='autograder education',
    url="https://github.com/nischalshrestha/pygradethis",
    packages=setuptools.find_packages(exclude=("tests")),
    install_requires=[
        "asttokens==2.0.4", 
        "numpy>=1.17.4",
        "pandas>=1.0.3", 
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Education",
    ],
    python_requires='>=3.6',
)
